@@ -26,6 +26,7 @@
 
 #include "pangocairo.h"
 #include "pangocairo-private.h"
+#include "pango-font-private.h"
 #include "pango-impl-utils.h"
 
 #define PANGO_CAIRO_FONT_PRIVATE(font)		\
@@ -92,7 +93,7 @@ done:
     {
       cairo_scaled_font_t *scaled_font = cf_priv->scaled_font;
       PangoFont *font = PANGO_FONT (cf_priv->cfont);
-      static GQuark warned_quark = 0;
+      static GQuark warned_quark = 0; /* MT-safe */
       if (!warned_quark)
 	warned_quark = g_quark_from_static_string ("pangocairo-scaledfont-warned");
 
@@ -136,17 +137,17 @@ done:
 
 /**
  * pango_cairo_font_get_scaled_font:
- * @font: a #PangoFont from a #PangoCairoFontMap
+ * @font: (nullable): a `PangoFont` from a `PangoCairoFontMap`
  *
- * Gets the #cairo_scaled_font_t used by @font.
+ * Gets the `cairo_scaled_font_t` used by @font.
  * The scaled font can be referenced and kept using
  * cairo_scaled_font_reference().
  *
- * Return value: the #cairo_scaled_font_t used by @font,
- *               or %NULL if @font is %NULL.
+ * Return value: (transfer none) (nullable): the `cairo_scaled_font_t`
+ *   used by @font
  *
  * Since: 1.18
- **/
+ */
 cairo_scaled_font_t *
 pango_cairo_font_get_scaled_font (PangoCairoFont *cfont)
 {
@@ -162,14 +163,14 @@ pango_cairo_font_get_scaled_font (PangoCairoFont *cfont)
 
 /**
  * _pango_cairo_font_install:
- * @font: a #PangoCairoFont
+ * @font: a `PangoCairoFont`
  * @cr: a #cairo_t
  *
  * Makes @font the current font for rendering in the specified
  * Cairo context.
  *
  * Return value: %TRUE if font was installed successfully, %FALSE otherwise.
- **/
+ */
 gboolean
 _pango_cairo_font_install (PangoFont *font,
 			   cairo_t   *cr)
@@ -223,6 +224,7 @@ _pango_cairo_font_get_metrics (PangoFont     *font,
   PangoCairoFontPrivate *cf_priv = PANGO_CAIRO_FONT_PRIVATE (font);
   PangoCairoFontMetricsInfo *info = NULL; /* Quiet gcc */
   GSList *tmp_list;
+  static int in_get_metrics;
 
   const char *sample_str = pango_language_get_sample_string (language);
 
@@ -246,9 +248,7 @@ _pango_cairo_font_get_metrics (PangoFont     *font,
       PangoRectangle extents;
       PangoFontDescription *desc;
       cairo_scaled_font_t *scaled_font;
-      cairo_matrix_t cairo_matrix;
-      PangoMatrix pango_matrix;
-      PangoMatrix identity = PANGO_MATRIX_INIT;
+      glong sample_str_width;
 
       int height, shift;
 
@@ -276,48 +276,32 @@ _pango_cairo_font_get_metrics (PangoFont     *font,
 
       info->metrics = (* PANGO_CAIRO_FONT_GET_IFACE (font)->create_base_metrics_for_context) (cfont, context);
 
-      /* We now need to adjust the base metrics for ctm */
-      cairo_scaled_font_get_ctm (scaled_font, &cairo_matrix);
-      pango_matrix.xx = cairo_matrix.xx;
-      pango_matrix.yx = cairo_matrix.yx;
-      pango_matrix.xy = cairo_matrix.xy;
-      pango_matrix.yy = cairo_matrix.yy;
-      pango_matrix.x0 = 0;
-      pango_matrix.y0 = 0;
-      if (G_UNLIKELY (0 != memcmp (&identity, &pango_matrix, 4 * sizeof (double))))
+      /* Ugly. We need to prevent recursion when we call into
+       * PangoLayout to determine approximate char width.
+       */
+      if (!in_get_metrics)
         {
-	  double xscale = pango_matrix_get_font_scale_factor (&pango_matrix);
-	  if (xscale) xscale = 1 / xscale;
+          in_get_metrics = 1;
 
-	  info->metrics->ascent *= xscale;
-	  info->metrics->descent *= xscale;
-	  info->metrics->underline_position *= xscale;
-	  info->metrics->underline_thickness *= xscale;
-	  info->metrics->strikethrough_position *= xscale;
-	  info->metrics->strikethrough_thickness *= xscale;
-	}
+          /* Update approximate_*_width now */
+          layout = pango_layout_new (context);
+          desc = pango_font_describe_with_absolute_size (font);
+          pango_layout_set_font_description (layout, desc);
+          pango_font_description_free (desc);
 
+          pango_layout_set_text (layout, sample_str, -1);
+          pango_layout_get_extents (layout, NULL, &extents);
 
-      /* Set the matrix on the context so we don't have to adjust the derived
-       * metrics. */
-      pango_context_set_matrix (context, &pango_matrix);
+          sample_str_width = pango_utf8_strwidth (sample_str);
+          g_assert (sample_str_width > 0);
+          info->metrics->approximate_char_width = extents.width / sample_str_width;
 
-      /* Update approximate_*_width now */
-      layout = pango_layout_new (context);
-      desc = pango_font_describe_with_absolute_size (font);
-      pango_layout_set_font_description (layout, desc);
-      pango_font_description_free (desc);
-
-      pango_layout_set_text (layout, sample_str, -1);
-      pango_layout_get_extents (layout, NULL, &extents);
-
-      info->metrics->approximate_char_width = extents.width / pango_utf8_strwidth (sample_str);
-
-      pango_layout_set_text (layout, "0123456789", -1);
-      info->metrics->approximate_digit_width = max_glyph_width (layout);
-
-      g_object_unref (layout);
+          pango_layout_set_text (layout, "0123456789", -1);
+          info->metrics->approximate_digit_width = max_glyph_width (layout);
 
+          g_object_unref (layout);
+          in_get_metrics = 0;
+        }
 
       /* We may actually reuse ascent/descent we got from cairo here.  that's
        * in cf_priv->font_extents.
@@ -357,7 +341,7 @@ _pango_cairo_font_get_metrics (PangoFont     *font,
 static PangoCairoFontHexBoxInfo *
 _pango_cairo_font_private_get_hex_box_info (PangoCairoFontPrivate *cf_priv)
 {
-  static const char hexdigits[] = "0123456789ABCDEF";
+  const char hexdigits[] = "0123456789ABCDEF";
   char c[2] = {0, 0};
   PangoFont *mini_font;
   PangoCairoFontHexBoxInfo *hbi;
@@ -376,9 +360,9 @@ _pango_cairo_font_private_get_hex_box_info (PangoCairoFontPrivate *cf_priv)
   double size, mini_size;
   PangoFontDescription *desc;
   cairo_scaled_font_t *scaled_font, *scaled_mini_font;
-  PangoMatrix pango_ctm;
-  cairo_matrix_t cairo_ctm;
-  PangoGravity gravity;
+  PangoMatrix pango_ctm, pango_font_matrix;
+  cairo_matrix_t cairo_ctm, cairo_font_matrix;
+  /*PangoGravity gravity;*/
 
   if (!cf_priv)
     return NULL;
@@ -394,10 +378,10 @@ _pango_cairo_font_private_get_hex_box_info (PangoCairoFontPrivate *cf_priv)
 
   font_options = cairo_font_options_create ();
   desc = pango_font_describe_with_absolute_size ((PangoFont *)cf_priv->cfont);
-  size = pango_font_description_get_size (desc) / (1.*PANGO_SCALE);
-  gravity = pango_font_description_get_gravity (desc);
+  /*gravity = pango_font_description_get_gravity (desc);*/
 
   cairo_scaled_font_get_ctm (scaled_font, &cairo_ctm);
+  cairo_scaled_font_get_font_matrix (scaled_font, &cairo_font_matrix);
   cairo_scaled_font_get_font_options (scaled_font, font_options);
   /* I started adding support for vertical hexboxes here, but it's too much
    * work.  Easier to do with cairo user fonts and vertical writing mode
@@ -410,8 +394,15 @@ _pango_cairo_font_private_get_hex_box_info (PangoCairoFontPrivate *cf_priv)
   pango_ctm.yy = cairo_ctm.yy;
   pango_ctm.x0 = cairo_ctm.x0;
   pango_ctm.y0 = cairo_ctm.y0;
+  pango_font_matrix.xx = cairo_font_matrix.xx;
+  pango_font_matrix.yx = cairo_font_matrix.yx;
+  pango_font_matrix.xy = cairo_font_matrix.xy;
+  pango_font_matrix.yy = cairo_font_matrix.yy;
+  pango_font_matrix.x0 = cairo_font_matrix.x0;
+  pango_font_matrix.y0 = cairo_font_matrix.y0;
 
-  size /= pango_matrix_get_font_scale_factor (&pango_ctm);
+  size = pango_matrix_get_font_scale_factor (&pango_font_matrix) /
+	 pango_matrix_get_font_scale_factor (&pango_ctm);
 
   if (is_hinted)
     {
@@ -459,13 +450,7 @@ _pango_cairo_font_private_get_hex_box_info (PangoCairoFontPrivate *cf_priv)
     mini_size = size / 2.2;
     if (is_hinted)
       {
-	mini_size = HINT_Y (mini_size);
-
-	if (mini_size < 6.0)
-	  {
-	    rows = 1;
-	    mini_size = MIN (MAX (size - 1, 0), 6.0);
-	  }
+        mini_size = HINT_Y (mini_size);
       }
 
     pango_font_description_set_absolute_size (desc, pango_units_from_double (mini_size));
@@ -486,8 +471,9 @@ _pango_cairo_font_private_get_hex_box_info (PangoCairoFontPrivate *cf_priv)
   pango_font_description_free (desc);
   cairo_font_options_destroy (font_options);
 
-
   scaled_mini_font = pango_cairo_font_get_scaled_font ((PangoCairoFont *) mini_font);
+  if (G_UNLIKELY (scaled_mini_font == NULL || cairo_scaled_font_status (scaled_mini_font) != CAIRO_STATUS_SUCCESS))
+    return NULL;
 
   for (i = 0 ; i < 16 ; i++)
     {
@@ -649,6 +635,28 @@ _pango_cairo_font_private_is_metrics_hinted (PangoCairoFontPrivate *cf_priv)
   return cf_priv->is_hinted;
 }
 
+
+static void
+get_space_extents (PangoCairoFontPrivate *cf_priv,
+                   PangoRectangle        *ink_rect,
+                   PangoRectangle        *logical_rect)
+{
+  /* See https://docs.microsoft.com/en-us/typography/develop/character-design-standards/whitespace */
+
+  int width = pango_font_get_absolute_size (PANGO_FONT (cf_priv->cfont)) / 4;
+
+  if (ink_rect)
+    {
+      ink_rect->x = ink_rect->y = ink_rect->height = 0;
+      ink_rect->width = width;
+    }
+  if (logical_rect)
+    {
+      *logical_rect = cf_priv->font_extents;
+      logical_rect->width = width;
+    }
+}
+
 static void
 _pango_cairo_font_private_get_glyph_extents_missing (PangoCairoFontPrivate *cf_priv,
 						     PangoGlyph             glyph,
@@ -659,6 +667,14 @@ _pango_cairo_font_private_get_glyph_extents_missing (PangoCairoFontPrivate *cf_p
   gunichar ch;
   gint rows, cols;
 
+  ch = glyph & ~PANGO_GLYPH_UNKNOWN_FLAG;
+
+  if (ch == 0x20 || ch == 0x2423)
+    {
+      get_space_extents (cf_priv, ink_rect, logical_rect);
+      return;
+    }
+
   hbi = _pango_cairo_font_private_get_hex_box_info (cf_priv);
   if (!hbi)
     {
@@ -666,19 +682,26 @@ _pango_cairo_font_private_get_glyph_extents_missing (PangoCairoFontPrivate *cf_p
       return;
     }
 
-  ch = glyph & ~PANGO_GLYPH_UNKNOWN_FLAG;
-
-  rows = hbi->rows;
   if (G_UNLIKELY (glyph == PANGO_GLYPH_INVALID_INPUT || ch > 0x10FFFF))
-    cols = 1;
+    {
+      rows = hbi->rows;
+      cols = 1;
+    }
+  else if (pango_get_ignorable_size (ch, &rows, &cols))
+    {
+      /* We special-case ignorables when rendering hex boxes */
+    }
   else
-    cols = ((glyph & ~PANGO_GLYPH_UNKNOWN_FLAG) > 0xffff ? 6 : 4) / rows;
+    {
+      rows = hbi->rows;
+      cols = ((glyph & ~PANGO_GLYPH_UNKNOWN_FLAG) > 0xffff ? 6 : 4) / rows;
+    }
 
   if (ink_rect)
     {
       ink_rect->x = PANGO_SCALE * hbi->pad_x;
       ink_rect->y = PANGO_SCALE * (hbi->box_descent - hbi->box_height);
-      ink_rect->width = PANGO_SCALE * (3 * hbi->pad_x + cols * (hbi->digit_width + hbi->pad_x));
+      ink_rect->width = PANGO_SCALE * (4 * hbi->pad_x + cols * (hbi->digit_width + hbi->pad_x));
       ink_rect->height = PANGO_SCALE * hbi->box_height;
     }
 
@@ -708,44 +731,65 @@ struct _PangoCairoFontGlyphExtentsCacheEntry
 static gboolean
 _pango_cairo_font_private_glyph_extents_cache_init (PangoCairoFontPrivate *cf_priv)
 {
-  cairo_scaled_font_t *scaled_font = _pango_cairo_font_private_get_scaled_font (cf_priv);
-  cairo_font_extents_t font_extents;
+  hb_font_extents_t extents;
 
-  if (G_UNLIKELY (scaled_font == NULL || cairo_scaled_font_status (scaled_font) != CAIRO_STATUS_SUCCESS))
-    return FALSE;
-
-  cairo_scaled_font_extents (scaled_font, &font_extents);
+  hb_font_get_h_extents (pango_font_get_hb_font (PANGO_FONT (cf_priv->cfont)),
+                         &extents);
 
   cf_priv->font_extents.x = 0;
   cf_priv->font_extents.width = 0;
-  cf_priv->font_extents.height = pango_units_from_double (font_extents.ascent + font_extents.descent);
+  cf_priv->font_extents.height = extents.ascender - extents.descender;
+
   switch (cf_priv->gravity)
     {
       default:
       case PANGO_GRAVITY_AUTO:
       case PANGO_GRAVITY_SOUTH:
-	cf_priv->font_extents.y = - pango_units_from_double (font_extents.ascent);
-	break;
+        cf_priv->font_extents.y = - extents.ascender;
+        break;
       case PANGO_GRAVITY_NORTH:
-	cf_priv->font_extents.y = - pango_units_from_double (font_extents.descent);
-	break;
+        cf_priv->font_extents.y = extents.descender;
+        break;
       case PANGO_GRAVITY_EAST:
       case PANGO_GRAVITY_WEST:
-	{
-	  int ascent = pango_units_from_double (font_extents.ascent + font_extents.descent) / 2;
-	  if (cf_priv->is_hinted)
-	    ascent = PANGO_UNITS_ROUND (ascent);
-	  cf_priv->font_extents.y = - ascent;
-	}
+        {
+          int ascent = cf_priv->font_extents.height / 2;
+          if (cf_priv->is_hinted)
+            ascent = PANGO_UNITS_ROUND (ascent);
+          cf_priv->font_extents.y = - ascent;
+        }
+       break;
     }
 
-  cf_priv->glyph_extents_cache = g_new0 (PangoCairoFontGlyphExtentsCacheEntry, GLYPH_CACHE_NUM_ENTRIES);
-  /* Make sure all cache entries are invalid initially */
-  cf_priv->glyph_extents_cache[0].glyph = 1; /* glyph 1 cannot happen in bucket 0 */
+  if (cf_priv->is_hinted)
+    {
+      if (cf_priv->font_extents.y < 0)
+        cf_priv->font_extents.y = PANGO_UNITS_FLOOR (cf_priv->font_extents.y);
+      else
+        cf_priv->font_extents.y = PANGO_UNITS_CEIL (cf_priv->font_extents.y);
+      if (cf_priv->font_extents.height < 0)
+        cf_priv->font_extents.height = PANGO_UNITS_FLOOR (extents.ascender) - PANGO_UNITS_CEIL (extents.descender);
+      else
+        cf_priv->font_extents.height = PANGO_UNITS_CEIL (extents.ascender) - PANGO_UNITS_FLOOR (extents.descender);
+    }
+
+  if (PANGO_GRAVITY_IS_IMPROPER (cf_priv->gravity))
+    {
+      cf_priv->font_extents.y = - cf_priv->font_extents.y;
+      cf_priv->font_extents.height = - cf_priv->font_extents.height;
+    }
+
+  if (!cf_priv->glyph_extents_cache)
+    {
+      cf_priv->glyph_extents_cache = g_new0 (PangoCairoFontGlyphExtentsCacheEntry, GLYPH_CACHE_NUM_ENTRIES);
+      /* Make sure all cache entries are invalid initially */
+      cf_priv->glyph_extents_cache[0].glyph = 1; /* glyph 1 cannot happen in bucket 0 */
+    }
 
   return TRUE;
 }
 
+
 /* Fills in the glyph extents cache entry
  */
 static void
@@ -764,7 +808,10 @@ compute_glyph_extents (PangoCairoFontPrivate  *cf_priv,
 				   &cairo_glyph, 1, &extents);
 
   entry->glyph = glyph;
-  entry->width = pango_units_from_double (extents.x_advance);
+  if (PANGO_GRAVITY_IS_VERTICAL (cf_priv->gravity))
+    entry->width = pango_units_from_double (extents.y_advance);
+  else
+    entry->width = pango_units_from_double (extents.x_advance);
   entry->ink_rect.x = pango_units_from_double (extents.x_bearing);
   entry->ink_rect.y = pango_units_from_double (extents.y_bearing);
   entry->ink_rect.width = pango_units_from_double (extents.width);
@@ -816,7 +863,7 @@ _pango_cairo_font_private_get_glyph_extents (PangoCairoFontPrivate *cf_priv,
     }
   else if (glyph & PANGO_GLYPH_UNKNOWN_FLAG)
     {
-      _pango_cairo_font_private_get_glyph_extents_missing(cf_priv, glyph, ink_rect, logical_rect);
+      _pango_cairo_font_private_get_glyph_extents_missing (cf_priv, glyph, ink_rect, logical_rect);
       return;
     }
 
@@ -827,6 +874,25 @@ _pango_cairo_font_private_get_glyph_extents (PangoCairoFontPrivate *cf_priv,
   if (logical_rect)
     {
       *logical_rect = cf_priv->font_extents;
-      logical_rect->width = entry->width;
+      switch (cf_priv->gravity)
+        {
+        case PANGO_GRAVITY_SOUTH:
+          logical_rect->width = entry->width;
+          break;
+        case PANGO_GRAVITY_EAST:
+          logical_rect->width = cf_priv->font_extents.height;
+          logical_rect->x = - logical_rect->width;
+          break;
+        case PANGO_GRAVITY_NORTH:
+          logical_rect->width = entry->width;
+          break;
+        case PANGO_GRAVITY_WEST:
+          logical_rect->width = - cf_priv->font_extents.height;
+          logical_rect->x = - logical_rect->width;
+          break;
+        case PANGO_GRAVITY_AUTO:
+        default:
+          g_assert_not_reached ();
+        }
     }
 }
